//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//
//  This is a listing of a program presented at the C++ at Work Conference.
//
//  The program was presented by William M. Miller of Glockenspiel, Inc.
//  in a seminar entitled Debugging C++. He stated that it was based on
//  an article by Martin J. O'Riordan, presented at the 1988 USENIX C++
//  conference.  (p227-242) in the proceedings.
//
//
//  Listings were provided for the Debugger but none were provided for the
//  example program.  I succeeded in copying freehand the example program
//  during the presentation.  
//
//  There was no mention of copyright on any of the distributed material
//  or within the listing.
//
//  It is with the spirit of learning OOP that I have typed this in and
//  debugged this version of this interesting program.  While it is not
//  anything close to the capability of the Turbo C++ IDE debugger or the
//  Turbo Debugger, I found it a unique and fresh approach to an
//  interesting problem.
//
//  Leland Berkwits  76414,430   January 8, 1991
//
//  Martin J. O'Riordan     martin@puschi.uucp
//  William M. Miller       wmmiller@cup.portal.com
//
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
//
//
// This routine implements the user interface of MOOD, the Minimal Object
// Oriented Debugger.  It has many limitations and is intended only as an
// illustration of what might be done.  In particular, it would be possible
// without much effort to allow symbolic lookup of object names by maintain
// ing a dynamic symbol table of live objects.
//
// In it's current form, MOOD recognizes the following commands:
//
// s           step to the next call of cond_break() and return to
//	       interactive mode
//
// g[<name>]   Go until the next call of cond_break() with the specified
//	       name (functio, object, or user breakpoint), or, if no name
//	       specified, until the next user breakpoint or the end of
//	       the program.
//
// d <addr>    Call the display() member function for the object whose
//	       "monitored" subobject is at the referenced address (the
//	       address is displayed at object construction time in verbose
//	       mode, and presumably display() member functions will do the
//	       same for linked monitored objects.
//
// v	       Turn on verbose mode (function entry and exit and object
//	       construction and destruction are echoed to stderr)
//
// q	       Turn off verbose mode.
//
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//

// Interface dependencies

#include <iostream.h>
#include <string.h>

#ifndef _DEBUGGER_DEFS
#define DEBUGGER_ON
#include "debugger.h"
#endif

// The objp() function does a system-dependent conversion of an ASCII
// pointer specification into a pointer to a monitored object.

monitored* objp(const char* str){ return (monitored *)str; }

// The cond_break() function is called under the various circumstances
// described by the enumeration break_condition.  It prints a message, if
// required, describing the reason for its call, and optionally enters
// interactive mode to take commands.

void cond_break( break_condition cond, const char* name, void* addr)
{
   static int tracing=0;		// => verbose mode
   static char brk_name[128]="";        // name on which to break
   static int was_step=1;		// last cmd was "step" => go to

   char buff[128];			// command line buffer

// We enter the display and possible interactive mode code under the
// following conditions:
//
// 1) We are tracing (verbose mode)
// 2) We are stepping.  (Note: this is initially TRUE, which takes care
//    of getting into interactive mode on the PROG_START call.)
// 3) The breakpoint name was set and the current name matches it, or
//    this is a user breakpoint call with no name.
// 4) The breakpoint name was not set and this is a user breakpoint call.
//

   if( tracing || was_step ||
       ( brk_name[0] && ((strcmp(brk_name, name)==0) ||
       ( cond == USER_BP && !name) ) ) ||
       (!brk_name[0] && cond == USER_BP))
   {
      switch(cond){
	 case PROG_START:
	   cerr<<"MOOD: Minimal Object Oriented Debugger, version0.0\n";
	   break;

	 case FCN_ENTRY:
	   cerr << "Enter " << name << "\n";
	   break;

	 case FCN_EXIT:
	   cerr << "Exit  " << name << "\n";
	   break;

	 case OBJ_CTOR:
	   cerr << "Construct " << name << " @ " << addr << "\n";
	   break;

	 case OBJ_DTOR:
	   cerr << "Destruct " << name << " @ " << addr << "\n";
	   break;

	 case USER_BP:
	   cerr << "Breakpoint " << name << " @ " << addr << "\n";
	   break;
      }

// We enter interactive mode if any of the above conditions other than
// tracing is met.  (This implies that named user breakpoints are skipped
// if the user uses a g <name> command in which the name does not match
// the breakpoint name, but that unnamed user breakpoints are always
// effective, as are named user breakpoints after a g command with no
// name.

      if( was_step || (brk_name[0] && ((strcmp(brk_name, name)==0) ||
	  (cond == USER_BP && !name))) ||
	  (!brk_name[0] && cond == USER_BP))
      {

	 was_step = 0;   // reset breakpoint conditions
	 brk_name[0]=0;

	 do{		 // main command loop
	    cerr << "cmd> " << flush;
	    cin >> buff;

	    switch(buff[0]){

	       case 'd':
		  objp(buff+2)->display();
		  break;

	       case 'g':     // go
		  if( buff[1])
		     strcpy(brk_name, buff+2);
		  break;

	       case 'q':     // quiet
		  tracing=0;
		  break;

	       case 's':     // single step
		  was_step=1;
		  break;

	       case 'v':
		  tracing=1;
		  break;
	    }                   // switch
	 }while( buff[0] != 's' && buff[0] != 'g');

      } // if interactive
   } // if message
} // cond_break()

