//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//
//  This is a listing of a program presented at the C++ at Work Conference.
//
//  The program was presented by William M. Miller of Glockenspiel, Inc.
//  in a seminar entitled Debugging C++. He stated that it was based on
//  an article by Martin J. O'Riordan, presented at the 1988 USENIX C++
//  conference.  (p227-242) in the proceedings.
//
//
//  Listings were provided for the Debugger but none were provided for the
//  example program.  I succeeded in copying freehand the example program
//  during the presentation.  
//
//  There was no mention of copyright on any of the distributed material
//  or within the listing.
//
//  It is with the spirit of learning OOP that I have typed this in and
//  debugged this version of this interesting program.  While it is not
//  anything close to the capability of the Turbo C++ IDE debugger or the
//  Turbo Debugger, I found it a unique and fresh approach to an
//  interesting problem.
//
//  Leland Berkwits  76414,430   January 8, 1991
//
//  Martin J. O'Riordan     martin@puschi.uucp
//  William M. Miller       wmmiller@cup.portal.com
//
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
//
// This header file is included in every module which is to be metered
// by the class debugger.  Its actions are controlled by the preprocessor
// variable DEBUGGER_ON:  if not defined, no debugger actions wil occur.

#ifndef _DEBUGGER_DEFS
#define _DEBUGGER_DEFS

// conditions under which cond_break may be called:

enum break_condition
{
     PROG_START,
     PROG_END,
     FCN_ENTRY,
     FCN_EXIT,
     OBJ_CTOR,
     OBJ_DTOR,
     USER_BP
};

// The cond_break() function is called in all the above contexts for two
// purposes:
//    1) To display trace information when verbose mode is on
//    2) To break under the appropriate conditions to allow the user to
//       interact with the debugger.  The default arguments are to allow
//       a user program to contain the call "cond_break();" with no
//	 arguments to perform an unconditional breakpoint into the
//       debugger's interactive mode.

void cond_break( break_condition cond=USER_BP,
		 const char* name =0,
		 void* addr = 0			);

// The trace class is intended to allow function tracing.  Each function
// or block which should be included in the trace should declare an object
// of class trace at the very beginning.  The result, in verbose mode, will
// be to display the function/block name at entry and exit;  this name can
// also be used to set a breakpoint from the debugger's interactive mode.

class trace{

#ifdef DEBUGGER_ON

private:
  const char* _name;
public:
   trace(const char *fcn_name): _name(fcn_name)
	{ cond_break(FCN_ENTRY, _name);  }

   ~trace(){ cond_break(FCN_EXIT, _name);  }

#else
public:
  trace(const char *){}
#endif
};

// The monitored class is intended for use as a virtual base class of
// any classes whose construction/destruction is to be traced in verbose
// mode or whose values are to be displayed interactively.  Dervied classes
// must pass the object or class name to the constructor and must supply an
// override to the display() member function.

class monitored{
public:

#ifdef DEBUGGER_ON
   monitored( const char* obj_name):_name(obj_name)
      { cond_break(OBJ_CTOR, _name, this);  }

   ~monitored(){ cond_break(OBJ_DTOR, _name, this); }

   virtual void display()=0;

private:
   const char* _name;

#else
   monitored(const char*){ }
#endif
};

// The following class and static object call cond_break exactly once at
// program start, before any debuggable object is created, to allow the
// user to set up tracing and breakpoints, and once at the end of execution
// to allow for any needed cleanup.

#ifdef DEBUGGER_ON

class init_ctl{
public:
   init_ctl()
   {   if(_count++ == 0)
	  cond_break(PROG_START);
   }

   ~init_ctl()
   {   if( --_count == 0)
	  cond_break(PROG_END);
   }

private:
   static int _count;
};

static init_ctl dbg_init;
#endif


#endif